import torch
import numpy as np
from matplotlib import pyplot as plt
from torchvision.utils import make_grid
import torchvision.transforms.functional as tf
import matplotlib.pyplot as plt

IMAGENET_DEFAULT_MEAN = (0.485, 0.456, 0.406)
IMAGENET_DEFAULT_STD = (0.229, 0.224, 0.225)

plt.rcParams['mathtext.fontset'] = 'cm'
plt.rcParams['font.family'] = 'serif'
plt.rcParams['font.size'] = 14
plt.rcParams["savefig.bbox"] = 'tight'

def show_grids(grids, titles, vertical=True):
    if vertical: 
        fig, axs = plt.subplots(len(grids), 1, figsize=(40, 10))
    else:
        fig, axs = plt.subplots(1, len(grids), figsize=(5 * len(grids), 5))
        fig.subplots_adjust(wspace=0.05)  # reduce horizontal space between images
    for ax, img, title in zip(axs, grids, titles):
        img = img.detach()
        img = tf.to_pil_image(img)
        ax.imshow(np.asarray(img))
        ax.set_title(title)
        ax.set(xticklabels=[], yticklabels=[], xticks=[], yticks=[])

def unnormalize(x):
    imagenet_mean_tensor = torch.tensor(IMAGENET_DEFAULT_MEAN).view(1,3,1,1).to(x.device, non_blocking=True)
    imagenet_std_tensor = torch.tensor(IMAGENET_DEFAULT_STD).view(1,3,1,1).to(x.device, non_blocking=True)
    x = torch.clip((x * imagenet_std_tensor + imagenet_mean_tensor) * 255, 0, 255).detach().cpu()/255
    return x 

def patchify(imgs, p):
    """
    imgs: (N, 3, H, W)
    x: (N, L, patch_size**2 *3)
    """
    # assert imgs.shape[2] == imgs.shape[3] and imgs.shape[2] % p == 0
    assert imgs.shape[2] % p == 0 and imgs.shape[3] % p == 0

    h, w = imgs.shape[2] // p, imgs.shape[3] // p 
    x = imgs.reshape(shape=(imgs.shape[0], 3, h, p, w, p))
    x = torch.einsum('nchpwq->nhwpqc', x)
    x = x.reshape(shape=(imgs.shape[0], h * w, p**2 * 3))
    return x

def unpatchify(x, patch_size, H, W, channels=3):
    """
    x: (N, L, patch_size**2 *channels)
    imgs: (N, 3, H, W)
    """
    p = patch_size
    h = H // p
    w = W // p
    assert h * w == x.shape[1]


    x = x.reshape(shape=(x.shape[0], h, w, p, p, channels))
    x = torch.einsum('nhwpqc->nchpwq', x)
    imgs = x.reshape(shape=(x.shape[0], channels, h * p, w * p))
    return imgs


def reconstruct_predicted_image(pred, seq, patchified, patch_size, mask, visible=True):
    B, S, C_in, H, W = seq.shape
    mean, var = patchified.mean(dim=-1, keepdim=True), patchified.var(dim=-1, keepdim=True)
    pred = pred.view(B*S, -1, patch_size**2*3)
    predicted_image = unpatchify(pred * (var + 1.e-6)**0.5 + mean, patch_size, H, W)
    if visible:
        # Replace visible patches with the original image
        image_masks = unpatchify(patchify(torch.ones_like(predicted_image), patch_size) * mask[:, :, None], patch_size, H, W)
        masked_target_image = (1 - image_masks) * seq.view(B*S, C_in, H, W)
        predicted_image = predicted_image * image_masks + masked_target_image
    predicted_image = predicted_image.view(B, S, C_in, H, W)
    return predicted_image

def qualitative_evaluation(model: torch.nn.Module, seq: torch.Tensor, path:str, visible:bool=True):
    patch_size = model.patch_size
    seq = seq[0].unsqueeze(0) # We only visualize the first batch element
    B, S, C_in, H, W = seq.shape
    with torch.inference_mode():
        loss, out, mask = model(seq)
        mask = mask.to(dtype=torch.bool)
        mask = mask.view(B*S, -1)
        patchified = patchify(seq.view(B*S, C_in, H, W), patch_size)        
        predicted_image = reconstruct_predicted_image(out, seq, patchified, patch_size, mask, visible=visible)

        masked_images = patchified.clone()
        masked_images[mask] = masked_images.min()
        masked_images = unpatchify(masked_images, patch_size, H, W)
        show_grids(
            [
                make_grid(unnormalize(seq[0]), normalize=True),
                make_grid(unnormalize(masked_images), normalize=True),
                make_grid(unnormalize(predicted_image[0]), normalize=True),
            ],
            [
                'Original',
                'Masked',
                'Predicted',
            ]
        )
        plt.show()
        plt.savefig(path)
        plt.close()


from argparse import Namespace
import os
import torch
import cv2
from time import time
from pathlib import Path
import matplotlib.cm as cm
import numpy as np

from src.models.topic_fm import TopicFM
from src import get_model_cfg
from .base import Viz
from src.utils.metrics import compute_symmetrical_epipolar_errors, compute_pose_errors
from src.utils.plotting import draw_topics, draw_topicfm_demo, error_colormap


class VizTopicFMv2(Viz):
    def __init__(self, args):
        super().__init__()
        if type(args) == dict:
            args = Namespace(**args)

        self.match_threshold = args.match_threshold
        # self.n_sampling_topics = args.n_sampling_topics
        self.show_n_topics = args.show_n_topics

        # Load model
        conf = dict(get_model_cfg())
        conf['match_coarse']['thr'] = self.match_threshold
        # conf['coarse']['n_samples'] = self.n_sampling_topics
        for k, v in args.coarse_model_cfg.items():
            conf["coarse"][k] = v
        conf['loss']['fine_type'] = "sym_epi"
        print("model config: ", conf)
        self.model = TopicFM(config=conf)
        ckpt_dict = torch.load(args.ckpt)
        self.model.load_state_dict(ckpt_dict['state_dict'])
        self.model = self.model.eval().to(self.device)

        # Name the method
        # self.ckpt_name = args.ckpt.split('/')[-1].split('.')[0]
        self.name = 'TopicFMv2'

        print(f'Initialize {self.name}')

    def match_and_draw(self, data_dict, root_dir=None, ground_truth=False, measure_time=False, viz_matches=True):
        if measure_time:
            torch.cuda.synchronize()
            start = torch.cuda.Event(enable_timing=True)
            end = torch.cuda.Event(enable_timing=True)
            start.record()
        self.model(data_dict)
        if measure_time:
            torch.cuda.synchronize()
            end.record()
            torch.cuda.synchronize()
            self.time_stats.append(start.elapsed_time(end))

            # self.flops_stats["backbone"].append(data_dict["backbone_flops"])
            # self.flops_stats["coarse_net"].append(data_dict["coarse_net_flops"])
            # self.flops_stats["fine_net"].append(data_dict["fine_net_flops"])
            # self.flops_stats["total"].append(data_dict["backbone_flops"] + data_dict["coarse_net_flops"] +
            #                                  data_dict["fine_net_flops"])


        kpts0 = data_dict['mkpts0_f'].cpu().numpy()
        kpts1 = data_dict['mkpts1_f'].cpu().numpy()

        img_name0, img_name1 = list(zip(*data_dict['pair_names']))[0]
        img0 = cv2.imread(os.path.join(root_dir, img_name0))
        img1 = cv2.imread(os.path.join(root_dir, img_name1))
        if str(data_dict["dataset_name"][0]).lower() == 'scannet':
            img0 = cv2.resize(img0, (640, 480))
            img1 = cv2.resize(img1, (640, 480))

        if viz_matches:
            saved_name = "_".join([img_name0.split('/')[-1].split('.')[0], img_name1.split('/')[-1].split('.')[0]])
            folder_matches = os.path.join(root_dir, "{}_viz_matches".format(self.name))
            if not os.path.exists(folder_matches):
                os.makedirs(folder_matches)
            path_to_save_matches = os.path.join(folder_matches, "{}.png".format(saved_name))

            if ground_truth:
                compute_symmetrical_epipolar_errors(data_dict)  # compute epi_errs for each match
                compute_pose_errors(data_dict, ransac_thr=0.4)  # compute R_errs, t_errs, pose_errs for each pair
                epi_errors = data_dict['epi_errs'].cpu().numpy()
                R_errors, t_errors = data_dict['R_errs'][0], data_dict['t_errs'][0]

                # self.draw_matches(kpts0, kpts1, img0, img1, epi_errors, path=path_to_save_matches,
                #                   R_errs=R_errors, t_errs=t_errors)

                # compute evaluation metrics
                rel_pair_names = list(zip(*data_dict['pair_names']))
                bs = data_dict['image0'].size(0)
                metrics = {
                    # to filter duplicate pairs caused by DistributedSampler
                    'identifiers': ['#'.join(rel_pair_names[b]) for b in range(bs)],
                    'epi_errs': [data_dict['epi_errs'][data_dict['m_bids'] == b].cpu().numpy() for b in range(bs)],
                    'R_errs': data_dict['R_errs'],
                    't_errs': data_dict['t_errs'],
                    'inliers': data_dict['inliers']}
                self.eval_stats.append({'metrics': metrics})
            else:
                m_conf = 1 - data_dict["mconf"].cpu().numpy()
                self.draw_matches(kpts0, kpts1, img0, img1, m_conf, path=path_to_save_matches, conf_thr=0.4)
            if self.show_n_topics > 0:
                folder_topics = os.path.join(root_dir, "{}_viz_topics".format(self.name))
                if not os.path.exists(folder_topics):
                    os.makedirs(folder_topics)
                draw_topics(data_dict, img0, img1, saved_folder=folder_topics, show_n_topics=self.show_n_topics,
                            saved_name=saved_name)

    def run_demo(self, dataloader, writer=None, output_dir=None, no_display=False, skip_frames=1):
        data_dict = next(dataloader)

        frame_id = 0
        last_image_id = 0
        img0 = np.array(cv2.imread(str(data_dict["img_path"][0])), dtype=np.float32) / 255
        frame_tensor = data_dict["img"].to(self.device)
        pair_data = {'image0': frame_tensor}
        last_frame = cv2.resize(img0, (frame_tensor.shape[-1], frame_tensor.shape[-2]), cv2.INTER_LINEAR)

        if output_dir is not None:
            print('==> Will write outputs to {}'.format(output_dir))
            Path(output_dir).mkdir(exist_ok=True)

        # Create a window to display the demo.
        if not no_display:
            window_name = 'Topic-assisted Feature Matching'
            cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
            cv2.resizeWindow(window_name, (640 * 2, 480 * 2))
        else:
            print('Skipping visualization, will not show a GUI.')

        # Print the keyboard help menu.
        print('==> Keyboard control:\n'
              '\tn: select the current frame as the reference image (left)\n'
              '\tq: quit')

        # vis_range = [kwargs["bottom_k"], kwargs["top_k"]]

        while True:
            frame_id += 1
            if frame_id == len(dataloader):
                print('Finished demo_loftr.py')
                break
            data_dict = next(dataloader)
            if frame_id % skip_frames != 0:
                # print("Skipping frame.")
                continue

            stem0, stem1 = last_image_id, data_dict["id"][0].item() - 1
            frame = np.array(cv2.imread(str(data_dict["img_path"][0])), dtype=np.float32) / 255

            frame_tensor = data_dict["img"].to(self.device)
            frame = cv2.resize(frame, (frame_tensor.shape[-1], frame_tensor.shape[-2]), interpolation=cv2.INTER_LINEAR)
            pair_data = {**pair_data, 'image1': frame_tensor}
            self.model(pair_data)

            total_n_matches = len(pair_data['mkpts0_f'])
            mkpts0 = pair_data['mkpts0_f'].cpu().numpy()  # [vis_range[0]:vis_range[1]]
            mkpts1 = pair_data['mkpts1_f'].cpu().numpy()  # [vis_range[0]:vis_range[1]]
            mconf = pair_data['mconf'].cpu().numpy()  # [vis_range[0]:vis_range[1]]

            # Normalize confidence.
            if len(mconf) > 0:
                mconf = 1 - mconf

            # alpha = 0
            # color = cm.jet(mconf, alpha=alpha)
            color = error_colormap(mconf, thr=0.4, alpha=0.1)

            text = [
                f'Topics',
                '#Matches: {}'.format(total_n_matches),
            ]

            out = draw_topicfm_demo(pair_data, last_frame, frame, mkpts0, mkpts1, color, text,
                                    show_n_topics=4, path=None)

            if not no_display:
                if writer is not None:
                    writer.write(out)
                cv2.imshow('TopicFM Matches', out)
                key = chr(cv2.waitKey(10) & 0xFF)
                if key == 'q':
                    if writer is not None:
                        writer.release()
                    print('Exiting...')
                    break
                elif key == 'n':
                    pair_data['image0'] = frame_tensor
                    last_frame = frame
                    last_image_id = (data_dict["id"][0].item() - 1)
                    frame_id_left = frame_id

            elif output_dir is not None:
                stem = 'matches_{:06}_{:06}'.format(stem0, stem1)
                out_file = str(Path(output_dir, stem + '.png'))
                print('\nWriting image to {}'.format(out_file))
                cv2.imwrite(out_file, out)
            else:
                raise ValueError("output_dir is required when no display is given.")

        cv2.destroyAllWindows()
        if writer is not None:
            writer.release()




import albumentations as A


class DarkAug(object):
    """
    Extreme dark augmentation aiming at Aachen Day-Night
    """

    def __init__(self) -> None:
        self.augmentor = A.Compose([
            A.RandomBrightnessContrast(p=0.75, brightness_limit=(-0.5, 0.0), contrast_limit=(-0.5, 0.3)),
            # A.Blur(p=0.1, blur_limit=(3, 9)),
            # A.MotionBlur(p=0.2, blur_limit=(3, 25)),
            # A.RandomGamma(p=0.1, gamma_limit=(15, 65)),
            A.HueSaturationValue(p=0.5, val_shift_limit=10)
        ], p=0.5)

    def __call__(self, x):
        return self.augmentor(image=x)['image']


class MobileAug(object):
    """
    Random augmentations aiming at images of mobile/handhold devices.
    """

    def __init__(self):
        self.augmentor = A.Compose([
            A.MotionBlur(p=0.25),
            A.ColorJitter(p=0.5),
            A.RandomRain(p=0.1),  # random occlusion
            A.RandomSunFlare(p=0.1),
            A.JpegCompression(p=0.25),
            A.ISONoise(p=0.25)
        ], p=1.0)

    def __call__(self, x):
        return self.augmentor(image=x)['image']


def build_augmentor(method=None, **kwargs):
    # if method is not None:
    #     raise NotImplementedError('Using of augmentation functions are not supported yet!')
    if method == 'dark':
        return DarkAug()
    elif method == 'mobile':
        return MobileAug()
    elif method is None:
        return None
    else:
        raise ValueError(f'Invalid augmentation method: {method}')


if __name__ == '__main__':
    augmentor = build_augmentor('FDA')



import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from kornia.geometry.subpix import dsnt
from kornia.utils.grid import create_meshgrid


class FineMatching(nn.Module):
    """FineMatching with s2d paradigm"""

    def __init__(self):
        super().__init__()

    def forward(self, feat_f0, feat_f1, data):
        """
        Args:
            feat0 (torch.Tensor): [M, WW, C]
            feat1 (torch.Tensor): [M, WW, C]
            data (dict)
        Update:
            data (dict):{
                'expec_f' (torch.Tensor): [M, 3],
                'mkpts0_f' (torch.Tensor): [M, 2],
                'mkpts1_f' (torch.Tensor): [M, 2]}
        """
        M, WW, C = feat_f0.shape
        W = int(math.sqrt(WW))
        scale = data['hw0_i'][0] / data['hw0_f'][0]
        self.M, self.W, self.WW, self.C, self.scale = M, W, WW, C, scale

        # corner case: if no coarse matches found
        if M == 0:
            assert self.training == False, "M is always >0, when training, see coarse_matching.py"
            # logger.warning('No matches found in coarse-level.')
            data.update({
                'expec_f': torch.empty(0, 3, device=feat_f0.device),
                'mkpts0_f': data['mkpts0_c'],
                'mkpts1_f': data['mkpts1_c'],
            })
            return

        feat_f0_picked = feat_f0[:, WW//2, :]

        sim_matrix = torch.einsum('mc,mrc->mr', feat_f0_picked, feat_f1)
        softmax_temp = 1. / C**.5
        heatmap = torch.softmax(softmax_temp * sim_matrix, dim=1)
        feat_f1_picked = (feat_f1 * heatmap.unsqueeze(-1)).sum(dim=1) # [M, C]
        heatmap = heatmap.view(-1, W, W)

        # compute coordinates from heatmap
        coords1_normalized = dsnt.spatial_expectation2d(heatmap[None], True)[0]  # [M, 2]
        grid_normalized = create_meshgrid(W, W, True, heatmap.device).reshape(1, -1, 2)  # [1, WW, 2]

        # compute std over <x, y>
        var = torch.sum(grid_normalized**2 * heatmap.view(-1, WW, 1), dim=1) - coords1_normalized**2  # [M, 2]
        std = torch.sum(torch.sqrt(torch.clamp(var, min=1e-10)), -1)  # [M]  clamp needed for numerical stability
        
        # for fine-level supervision
        data.update({'expec_f': torch.cat([coords1_normalized, std.unsqueeze(1)], -1),
                     'descriptors0': feat_f0_picked.detach(), 'descriptors1': feat_f1_picked.detach()})

        # compute absolute kpt coords
        self.get_fine_match(coords1_normalized, data)

    @torch.no_grad()
    def get_fine_match(self, coords1_normed, data):
        W, WW, C, scale = self.W, self.WW, self.C, self.scale

        # mkpts0_f and mkpts1_f
        # scale0 = scale * data['scale0'][data['b_ids']] if 'scale0' in data else scale
        mkpts0_f = data['mkpts0_c'] # + (coords0_normed * (W // 2) * scale0 )[:len(data['mconf'])]
        scale1 = scale * data['scale1'][data['b_ids']] if 'scale1' in data else scale
        mkpts1_f = data['mkpts1_c'] + (coords1_normed * (W // 2) * scale1)[:len(data['mconf'])]

        data.update({
            "mkpts0_f": mkpts0_f,
            "mkpts1_f": mkpts1_f
        })


class DynamicFineMatching(nn.Module):
    """FineMatching with s2d paradigm"""

    def __init__(self):
        super().__init__()

    def forward(self, feat_f0, feat_f1, data):
        """
        Args:
            feat0 (torch.Tensor): [M, WW, C]
            feat1 (torch.Tensor): [M, WW, C]
            data (dict)
        Update:
            data (dict):{
                'expec_f' (torch.Tensor): [M, 3],
                'mkpts0_f' (torch.Tensor): [M, 2],
                'mkpts1_f' (torch.Tensor): [M, 2]}
        """
        M, WW, C = feat_f0.shape
        W = int(math.sqrt(WW))
        scale = data['hw0_i'][0] / data['hw0_f'][0]
        self.M, self.W, self.WW, self.C= M, W, WW, C
        self.scale0 = scale * data['scale0'][data['b_ids']] if 'scale0' in data else scale
        self.scale1 = scale * data['scale1'][data['b_ids']] if 'scale1' in data else scale
        # grid_WW = create_meshgrid(W, W, normalized_coordinates=True, device=feat_f0.device).reshape(1, -1, 2).repeat(M, 1, 1)

        # corner case: if no coarse matches found
        if M == 0:
            assert self.training == False, "M is always >0, when training, see coarse_matching.py"
            # logger.warning('No matches found in coarse-level.')
            data.update({
                'f_b_ids': data['b_ids'],
                'all_mkpts0_f': data['all_mkpts0_c'],
                'all_mkpts1_f': data['all_mkpts1_c'],
                'mkpts0_f': data['mkpts0_c'],
                'mkpts1_f': data['mkpts1_c'],
            })
            return

        temperature = 0.1
        score_map0 = data["score_map0"]
        if self.training:
            temperature = math.pow(2, -0.5*data["epoch_idx"]) if data["epoch_idx"] < 8 else 0.1
        heatmap0 = F.softmax(score_map0 / temperature, dim=-1)
        feat_f0_picked = (feat_f0 * heatmap0.unsqueeze(-1)).sum(dim=1)
        coords0_normed = dsnt.spatial_expectation2d(heatmap0.view(-1, W, W)[None], True)[0]
        out = self.get_fine_match(data, coords0_normed, feat_f0_picked, feat_f1)
        out["mconf"] *= data["mconf"]
        data.update(out)

    def get_fine_match(self, data, coords0_normed, feat_f0_picked, feat_f1, mask_b_ids=None):
        if mask_b_ids is None:
            mask_b_ids = torch.ones_like(data["b_ids"]).bool()
        coords0_normed = coords0_normed[mask_b_ids]
        feat_f0_picked = feat_f0_picked[mask_b_ids]
        feat_f1 = feat_f1[mask_b_ids]

        sim_matrix = torch.einsum('mc,mrc->mr', feat_f0_picked, feat_f1) / self.C ** .5
        heatmap1 = torch.softmax(sim_matrix, dim=1)
        mconf, _ = torch.max(heatmap1, dim=1)
        # data.update({"heatmap1": heatmap1})
        # feat_f1_picked = (feat_f1 * heatmap1.unsqueeze(-1)).sum(dim=1)  # [M, C]
        heatmap1 = heatmap1.view(-1, self.W, self.W)

        # compute coordinates from heatmap
        coords1_normed = dsnt.spatial_expectation2d(heatmap1[None], True)[0]
        scale0 = self.scale0[mask_b_ids] if isinstance(self.scale0, torch.Tensor) else self.scale0
        all_mkpts0_f = data['all_mkpts0_c'][mask_b_ids] + (coords0_normed * (self.W // 2) * scale0)
        scale1 = self.scale1[mask_b_ids] if isinstance(self.scale1, torch.Tensor) else self.scale1
        all_mkpts1_f = data['all_mkpts1_c'][mask_b_ids] + (coords1_normed * (self.W // 2) * scale1)

        # mkpts0_f and mkpts1_f
        true_matches = (~data["gt_mask"])[mask_b_ids]
        mkpts0_f = all_mkpts0_f.detach()[true_matches]
        mkpts1_f = all_mkpts1_f.detach()[true_matches]
        scale0_f = scale0[true_matches] if isinstance(scale0, torch.Tensor) else scale0
        scale1_f = scale1[true_matches] if isinstance(scale1, torch.Tensor) else scale1
        mconf = mconf.detach()[true_matches]
        f_b_ids = data["b_ids"][mask_b_ids]

        return {
            "f_b_ids": f_b_ids,
            "m_bids": f_b_ids[true_matches],
            "mconf": mconf,
            "all_mkpts0_f": all_mkpts0_f,
            "all_mkpts1_f": all_mkpts1_f,
            "mkpts0_f": mkpts0_f,
            "mkpts1_f": mkpts1_f,
            "scale0_f": scale0_f, "scale1_f": scale1_f
        }



import os
from typing import Union
from loguru import logger
from itertools import chain

import torch
from yacs.config import CfgNode as CN


def lower_config(yacs_cfg):
    if not isinstance(yacs_cfg, CN):
        return yacs_cfg
    return {k.lower(): lower_config(v) for k, v in yacs_cfg.items()}


def upper_config(dict_cfg):
    if not isinstance(dict_cfg, dict):
        return dict_cfg
    return {k.upper(): upper_config(v) for k, v in dict_cfg.items()}


def log_on(condition, message, level):
    if condition:
        assert level in ['INFO', 'DEBUG', 'WARNING', 'ERROR', 'CRITICAL']
        logger.log(level, message)


def setup_gpus(gpus: Union[str, int]) -> int:
    """ A temporary fix for pytorch-lighting 1.3.x """
    gpus = str(gpus)
    gpu_ids = []

    if ',' not in gpus:
        n_gpus = int(gpus)
        return n_gpus if n_gpus != -1 else torch.cuda.device_count()
    else:
        gpu_ids = [i.strip() for i in gpus.split(',') if i != '']

    # setup environment variables
    visible_devices = os.getenv('CUDA_VISIBLE_DEVICES')
    if visible_devices is None:
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = ','.join(str(i) for i in gpu_ids)
        visible_devices = os.getenv('CUDA_VISIBLE_DEVICES')
        logger.warning(
            f'[Temporary Fix] manually set CUDA_VISIBLE_DEVICES when specifying gpus to use: {visible_devices}')
    else:
        logger.warning(
            '[Temporary Fix] CUDA_VISIBLE_DEVICES already set by user or the main process.')
    return len(gpu_ids)


def flattenList(x):
    return list(chain(*x))



from .attention import CausalSelfAttention, LinearKMaskedBias, SelfAttention
from .block import CausalSelfAttentionBlock, SelfAttentionBlock
from .ffn_layers import Mlp
from .layer_scale import LayerScale
from .patch_embed import PatchEmbed
from .rope_position_encoding import RopePositionEmbedding


import os
import random
import subprocess
from typing import Callable, List, Optional, Tuple
import numpy as np
import torch
from torch import Tensor, nn

def cat_keep_shapes(x_list: List[Tensor]) -> Tuple[Tensor, List[Tuple[int]], List[int]]:
    shapes = [x.shape for x in x_list]
    num_tokens = [x.select(dim=-1, index=0).numel() for x in x_list]
    flattened = torch.cat([x.flatten(0, -2) for x in x_list])
    return flattened, shapes, num_tokens


def uncat_with_shapes(flattened: Tensor, shapes: List[Tuple[int]], num_tokens: List[int]) -> List[Tensor]:
    outputs_splitted = torch.split_with_sizes(flattened, num_tokens, dim=0)
    shapes_adjusted = [shape[:-1] + torch.Size([flattened.shape[-1]]) for shape in shapes]
    outputs_reshaped = [o.reshape(shape) for o, shape in zip(outputs_splitted, shapes_adjusted)]
    return outputs_reshaped

def named_replace(
    fn: Callable,
    module: nn.Module,
    name: str = "",
    depth_first: bool = True,
    include_root: bool = False,
) -> nn.Module:
    if not depth_first and include_root:
        module = fn(module=module, name=name)
    for child_name_o, child_module in list(module.named_children()):
        child_name = ".".join((name, child_name_o)) if name else child_name_o
        new_child = named_replace(
            fn=fn,
            module=child_module,
            name=child_name,
            depth_first=depth_first,
            include_root=True,
        )
        setattr(module, child_name_o, new_child)

    if depth_first and include_root:
        module = fn(module=module, name=name)
    return module


def named_apply(
    fn: Callable,
    module: nn.Module,
    name: str = "",
    depth_first: bool = True,
    include_root: bool = False,
) -> nn.Module:
    if not depth_first and include_root:
        fn(module=module, name=name)
    for child_name, child_module in module.named_children():
        child_name = ".".join((name, child_name)) if name else child_name
        named_apply(
            fn=fn,
            module=child_module,
            name=child_name,
            depth_first=depth_first,
            include_root=True,
        )
    if depth_first and include_root:
        fn(module=module, name=name)
    return module


def fix_random_seeds(seed: int = 31):
    """
    Fix random seeds.
    """
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def get_sha() -> str:
    cwd = os.path.dirname(os.path.abspath(__file__))

    def _run(command):
        return subprocess.check_output(command, cwd=cwd).decode("ascii").strip()

    sha = "N/A"
    diff = "clean"
    branch = "N/A"
    try:
        sha = _run(["git", "rev-parse", "HEAD"])
        subprocess.check_output(["git", "diff"], cwd=cwd)
        diff = _run(["git", "diff-index", "HEAD"])
        diff = "has uncommited changes" if diff else "clean"
        branch = _run(["git", "rev-parse", "--abbrev-ref", "HEAD"])
    except Exception:
        pass
    message = f"sha: {sha}, status: {diff}, branch: {branch}"
    return message


def get_conda_env() -> Tuple[Optional[str], Optional[str]]:
    conda_env_name = os.environ.get("CONDA_DEFAULT_ENV")
    conda_env_path = os.environ.get("CONDA_PREFIX")
    return conda_env_name, conda_env_path


def count_parameters(module: nn.Module) -> int:
    c = 0
    for m in module.parameters():
        c += m.nelement()
    return c


def has_batchnorms(model: nn.Module) -> bool:
    bn_types = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm)
    for _, module in model.named_modules():
        if isinstance(module, bn_types):
            return True
    return False


import torch
import torch.nn as nn
from enum import Enum
from .vit_model import mum_vitl16
from types import MethodType

class FwdType(str, Enum):
    SINGLE_VIEW = 'single_view'
    MATCHING = 'matching'

def custom_fwd_single_view(
    self: nn.Module,
    x: torch.Tensor,
):
    bs, _, h, w = x.shape
    x = self.forward_encoder(x, 0, return_all_blocks=True)[-1]
    x = self.norm(x)
    outcome = x[:, 0]
    patch_h, patch_w = self.patch_embed.patch_size
    return outcome, None, x[:, 1:].reshape(bs, h // patch_h, w // patch_w, -1)

def custom_fwd_matching(
    self: nn.Module,
    x1: torch.Tensor,
    x2: torch.Tensor,
):
    feat1 = self.forward_features(x1)['x_norm_patchtokens']
    feat2 = self.forward_features(x2)['x_norm_patchtokens']
    return feat1, feat2
    # feat1 = self.forward_encoder(x1, 0, return_all_blocks=True)[-1]
    # feat2 = self.forward_encoder(x2, 0, return_all_blocks=True)[-1]
    # return feat1[:, 1:], feat2[:, 1:]  # Exclude cls token

def __model_loader__(
    fwd: FwdType = FwdType.MATCHING
) -> nn.Module: 

    model = mum_vitl16(pretrained=True)

    if fwd == FwdType.SINGLE_VIEW:
        model.forward = MethodType(custom_fwd_single_view, model)
    elif fwd == FwdType.MATCHING:
        model.forward = MethodType(custom_fwd_matching, model)
    else:
        raise ValueError(f"Unknown forward type: {fwd}")

    model.cuda()
    model.eval()
    return model
    